{"cells": [{"cell_type": "markdown", "metadata": {"tags": ["module-dsml"]}, "source": ["(Coding_Example)=\n", "# Chapter 20 -- Coding Example\n", "[<!-- module-dsml badge --><span class=\"module module-dsml\">Data Science and Machine Learning for Geoscientists</span>](module-dsml) \n"]}, {"cell_type": "code", "execution_count": 45, "metadata": {}, "outputs": [], "source": ["import numpy as np\n", "import random\n", "class Network(object):\n", "\n", "    def __init__(self, sizes):\n", "        self.num_layers = len(sizes)    # sizes = [2,3,1] --> it has 2 inputs, 3 neurons in the second layer, and 1 neuron in the last layer\n", "        self.sizes = sizes    # len of sizes is how many layers it has, it this case, 3 layers\n", "        self.biases = [np.random.randn(y, 1) for y in sizes[1:]]    # y = 3 and 1\n", "        self.weights = [np.random.randn(y, x)     # np.random.randn(y, 1)  creates an array containing y lists, each has 1 random value\n", "                        for x, y in zip(sizes[:-1], sizes[1:])]    # np.random.randn mean 0 and variance 1\n", "                                        # 2,3          3,1\n", "    def feedforward(self, a):\n", "        \"\"\"Return the output of the network if \"a\" is input.\"\"\"\n", "        for b, w in zip(self.biases, self.weights):\n", "            a = sigmoid(np.dot(w, a)+b)\n", "        return a    # a is an array that contains\n", "    \n", "    def SGD(self, training_data, epochs, mini_batch_size, eta,\n", "            test_data=None):\n", "        \"\"\"Train the neural network using mini-batch stochastic\n", "        gradient descent.  The \"training_data\" is a list of tuples\n", "        \"(x, y)\" representing the training inputs and the desired\n", "        outputs.  The other non-optional parameters are\n", "        self-explanatory.  If \"test_data\" is provided then the\n", "        network will be evaluated against the test data after each\n", "        epoch, and partial progress printed out.  This is useful for\n", "        tracking progress, but slows things down substantially.\"\"\"\n", "        if test_data: n_test = len(test_data)\n", "        n = len(training_data)\n", "        for j in range(epochs):\n", "            random.shuffle(training_data)\n", "            mini_batches = [\n", "                training_data[k:k+mini_batch_size]\n", "                for k in range(0, n, mini_batch_size)]\n", "            for mini_batch in mini_batches:\n", "                self.update_mini_batch(mini_batch, eta)\n", "            if test_data:\n", "                print(\"Epoch {0}: {1} / {2}\".format(\n", "                    j, self.evaluate(test_data), n_test))\n", "            else:\n", "                print(\"Epoch {0} complete\".format(j))\n", "                \n", "    def update_mini_batch(self, mini_batch, eta):\n", "        \"\"\"Update the network's weights and biases by applying\n", "        gradient descent using backpropagation to a single mini batch.\n", "        The \"mini_batch\" is a list of tuples \"(x, y)\", and \"eta\"\n", "        is the learning rate.\"\"\"\n", "        nabla_b = [np.zeros(b.shape) for b in self.biases]\n", "        nabla_w = [np.zeros(w.shape) for w in self.weights]\n", "        for x, y in mini_batch:\n", "            delta_nabla_b, delta_nabla_w = self.backprop(x, y)\n", "            nabla_b = [nb+dnb for nb, dnb in zip(nabla_b, delta_nabla_b)]\n", "            nabla_w = [nw+dnw for nw, dnw in zip(nabla_w, delta_nabla_w)]\n", "        self.weights = [w-(eta/len(mini_batch))*nw \n", "                        for w, nw in zip(self.weights, nabla_w)]\n", "        self.biases = [b-(eta/len(mini_batch))*nb \n", "                       for b, nb in zip(self.biases, nabla_b)]\n", "        \n", "    def backprop(self, x, y):\n", "        \"\"\"Return a tuple ``(nabla_b, nabla_w)`` representing the\n", "        gradient for the cost function C_x.  ``nabla_b`` and\n", "        ``nabla_w`` are layer-by-layer lists of numpy arrays, similar\n", "        to ``self.biases`` and ``self.weights``.\"\"\"\n", "        nabla_b = [np.zeros(b.shape) for b in self.biases]\n", "        nabla_w = [np.zeros(w.shape) for w in self.weights]\n", "        # feedforward\n", "        activation = x\n", "        activations = [x] # list to store all the activations, layer by layer\n", "        zs = [] # list to store all the z vectors, layer by layer\n", "        for b, w in zip(self.biases, self.weights):\n", "            z = np.dot(w, activation)+b\n", "            zs.append(z)\n", "            activation = sigmoid(z)\n", "            activations.append(activation)\n", "        # backward pass\n", "        delta = self.cost_derivative(activations[-1], y) * sigmoid_prime(zs[-1])\n", "        nabla_b[-1] = delta\n", "        nabla_w[-1] = np.dot(delta, activations[-2].transpose())\n", "        # Note that the variable l in the loop below is used a little\n", "        # differently to the notation in Chapter 2 of the book.  Here,\n", "        # l = 1 means the last layer of neurons, l = 2 is the\n", "        # second-last layer, and so on.  It's a renumbering of the\n", "        # scheme in the book, used here to take advantage of the fact\n", "        # that Python can use negative indices in lists.\n", "        for l in range(2, self.num_layers):\n", "            z = zs[-l]\n", "            sp = sigmoid_prime(z)\n", "            delta = np.dot(self.weights[-l+1].transpose(), delta) * sp\n", "            nabla_b[-l] = delta\n", "            nabla_w[-l] = np.dot(delta, activations[-l-1].transpose())\n", "        return (nabla_b, nabla_w)\n", "    \n", "    def evaluate(self, test_data):\n", "        \"\"\"Return the number of test inputs for which the neural\n", "        network outputs the correct result. Note that the neural\n", "        network's output is assumed to be the index of whichever\n", "        neuron in the final layer has the highest activation.\"\"\"\n", "        test_results = [(np.argmax(self.feedforward(x)), y)\n", "                        for (x, y) in test_data]\n", "        return sum(int(x == y) for (x, y) in test_results)\n", "\n", "    def cost_derivative(self, output_activations, y):\n", "        \"\"\"Return the vector of partial derivatives \\partial C_x /\n", "        \\partial a for the output activations.\"\"\"\n", "        return (output_activations-y)\n"]}, {"cell_type": "code", "execution_count": 46, "metadata": {}, "outputs": [], "source": ["def sigmoid(z):\n", "        return 1.0/(1.0+np.exp(-z))\n", "def sigmoid_prime(z):\n", "    \"\"\"Derivative of the sigmoid function.\"\"\"\n", "    return sigmoid(z)*(1-sigmoid(z))"]}, {"cell_type": "code", "execution_count": 47, "metadata": {}, "outputs": [{"data": {"text/plain": ["array([[ 2.06118074],\n", "       [-0.42243013],\n", "       [-0.89537547]])"]}, "execution_count": 47, "metadata": {}, "output_type": "execute_result"}], "source": ["sizes = [2, 3, 1]\n", "for y in sizes[1:]:\n", "    #print(y)\n", "    pass\n", "np.random.randn(3, 1)"]}, {"cell_type": "code", "execution_count": 48, "metadata": {}, "outputs": [], "source": ["net = Network([2, 3, 1])"]}, {"cell_type": "code", "execution_count": 1, "metadata": {}, "outputs": [], "source": ["\"\"\"\n", "mnist_loader\n", "~~~~~~~~~~~~\n", "\n", "A library to load the MNIST image data.  For details of the data\n", "structures that are returned, see the doc strings for ``load_data``\n", "and ``load_data_wrapper``.  In practice, ``load_data_wrapper`` is the\n", "function usually called by our neural network code.\n", "\"\"\"\n", "\n", "#### Libraries\n", "# Standard library\n", "import pickle\n", "import gzip\n", "\n", "# Third-party libraries\n", "import numpy as np\n", "\n", "def load_data():\n", "    \"\"\"Return the MNIST data as a tuple containing the training data,\n", "    the validation data, and the test data.\n", "\n", "    The ``training_data`` is returned as a tuple with two entries.\n", "    The first entry contains the actual training images.  This is a\n", "    numpy ndarray with 50,000 entries.  Each entry is, in turn, a\n", "    numpy ndarray with 784 values, representing the 28 * 28 = 784\n", "    pixels in a single MNIST image.\n", "\n", "    The second entry in the ``training_data`` tuple is a numpy ndarray\n", "    containing 50,000 entries.  Those entries are just the digit\n", "    values (0...9) for the corresponding images contained in the first\n", "    entry of the tuple.\n", "\n", "    The ``validation_data`` and ``test_data`` are similar, except\n", "    each contains only 10,000 images.\n", "\n", "    This is a nice data format, but for use in neural networks it's\n", "    helpful to modify the format of the ``training_data`` a little.\n", "    That's done in the wrapper function ``load_data_wrapper()``, see\n", "    below.\n", "    \"\"\"\n", "    f = gzip.open('mnist.pkl.gz', 'rb')\n", "    training_data, validation_data, test_data = pickle.load(f, encoding='latin1')\n", "    f.close()\n", "    return (training_data, validation_data, test_data)\n", "\n", "def load_data_wrapper():\n", "    \"\"\"Return a tuple containing ``(training_data, validation_data,\n", "    test_data)``. Based on ``load_data``, but the format is more\n", "    convenient for use in our implementation of neural networks.\n", "\n", "    In particular, ``training_data`` is a list containing 50,000\n", "    2-tuples ``(x, y)``.  ``x`` is a 784-dimensional numpy.ndarray\n", "    containing the input image.  ``y`` is a 10-dimensional\n", "    numpy.ndarray representing the unit vector corresponding to the\n", "    correct digit for ``x``.\n", "\n", "    ``validation_data`` and ``test_data`` are lists containing 10,000\n", "    2-tuples ``(x, y)``.  In each case, ``x`` is a 784-dimensional\n", "    numpy.ndarry containing the input image, and ``y`` is the\n", "    corresponding classification, i.e., the digit values (integers)\n", "    corresponding to ``x``.\n", "\n", "    Obviously, this means we're using slightly different formats for\n", "    the training data and the validation / test data.  These formats\n", "    turn out to be the most convenient for use in our neural network\n", "    code.\"\"\"\n", "    tr_d, va_d, te_d = load_data()\n", "    training_inputs = [np.reshape(x, (784, 1)) for x in tr_d[0]]\n", "    training_results = [vectorized_result(y) for y in tr_d[1]]\n", "    training_data = list(zip(training_inputs, training_results)) \n", "    validation_inputs = [np.reshape(x, (784, 1)) for x in va_d[0]]\n", "    validation_data = list(zip(validation_inputs, va_d[1]))\n", "    test_inputs = [np.reshape(x, (784, 1)) for x in te_d[0]]\n", "    test_data = list(zip(test_inputs, te_d[1])) \n", "    return (training_data, validation_data, test_data)\n", "\n", "def vectorized_result(j):\n", "    \"\"\"Return a 10-dimensional unit vector with a 1.0 in the jth\n", "    position and zeroes elsewhere.  This is used to convert a digit\n", "    (0...9) into a corresponding desired output from the neural\n", "    network.\"\"\"\n", "    e = np.zeros((10, 1))\n", "    e[j] = 1.0\n", "    return e"]}, {"cell_type": "code", "execution_count": 50, "metadata": {}, "outputs": [], "source": ["training_data, validation_data, test_data = load_data_wrapper()"]}, {"cell_type": "code", "execution_count": 51, "metadata": {}, "outputs": [{"name": "stdout", "output_type": "stream", "text": ["Epoch 0: 8281 / 10000\n", "Epoch 1: 8372 / 10000\n", "Epoch 2: 9320 / 10000\n", "Epoch 3: 9359 / 10000\n", "Epoch 4: 9395 / 10000\n", "Epoch 5: 9395 / 10000\n", "Epoch 6: 9416 / 10000\n", "Epoch 7: 9437 / 10000\n", "Epoch 8: 9469 / 10000\n", "Epoch 9: 9444 / 10000\n", "Epoch 10: 9464 / 10000\n", "Epoch 11: 9471 / 10000\n", "Epoch 12: 9450 / 10000\n", "Epoch 13: 9488 / 10000\n", "Epoch 14: 9478 / 10000\n", "Epoch 15: 9465 / 10000\n", "Epoch 16: 9476 / 10000\n", "Epoch 17: 9482 / 10000\n", "Epoch 18: 9490 / 10000\n", "Epoch 19: 9486 / 10000\n", "Epoch 20: 9505 / 10000\n", "Epoch 21: 9497 / 10000\n", "Epoch 22: 9496 / 10000\n", "Epoch 23: 9491 / 10000\n", "Epoch 24: 9494 / 10000\n", "Epoch 25: 9508 / 10000\n", "Epoch 26: 9506 / 10000\n", "Epoch 27: 9496 / 10000\n", "Epoch 28: 9493 / 10000\n", "Epoch 29: 9518 / 10000\n"]}], "source": ["net = Network([784, 30, 10])\n", "net.SGD(training_data, 30, 10, 3.0, test_data=test_data)"]}], "metadata": {"celltoolbar": "Tags", "kernelspec": {"display_name": "Python 3", "language": "python", "name": "python3"}, "language_info": {"codemirror_mode": {"name": "ipython", "version": 3}, "file_extension": ".py", "mimetype": "text/x-python", "name": "python", "nbconvert_exporter": "python", "pygments_lexer": "ipython3", "version": "3.8.8"}}, "nbformat": 4, "nbformat_minor": 4}